@@ -2,13 +2,14 @@ package Date::ICal;
 use strict;
 
 use vars qw($VERSION);
-$VERSION = (qw'$Revision: 1.15 $')[1];
+$VERSION = (qw'$Revision: 1.20 $')[1];
 use Carp;
 use Time::Local;
+use Date::Leapyear qw();
 
 =head1 NAME
 
-Date::ICal - Perl extension for ICal date objects.
+Date::ICal - Perl extension for ICalendar date objects.
 
 =head1 SYNOPSIS
 
@@ -16,6 +17,9 @@ Date::ICal - Perl extension for ICal date objects.
 
     $ical = Date::ICal->new( ical => '19971024T120000' );
     $ical = Date::ICal->new( epoch => time );
+    $ical = Date::ICal->new( year => 1964,
+        month => 10, day => 16, hour => 16,
+        min => 12, sec => 47, tz => '0530' );
 
     $hour = $ical->hour;
     $year = $ical->year;
@@ -23,8 +27,6 @@ Date::ICal - Perl extension for ICal date objects.
     $ical_string = $ical->ical;
     $epoch_time = $ical->epoch;
 
-    $ical->epoch( time+60 ); # Set the time a minute ahead
-
 =head1 DESCRIPTION
 
 Date::ICal talks the ICal date format, and is intended to be a base class for
@@ -46,6 +48,17 @@ Or with any epoch time:
 
     my $ical = Date::ICal->new( epoch => time );
 
+Or, better still, create it with components
+
+    my $date = Date::ICal->new( 
+                           day => 25, 
+                           month => 10, 
+                           year => 1066,
+                           hour => 7,
+                           min => 15,
+                           sec => 47
+                           );
+
 If you call new without any arguments, you'll get a Date::ICal object that is
 set to the time right now.
 
@@ -58,11 +71,13 @@ use Date::ICal;
 
 my $t1 = Date::ICal->new(epoch => '0');
 ok ($t1->epoch() eq '0', 'creation test from epoch (compare to epoch)');
-ok ($t1->ical() eq '19700101', 'creation test from epoch (compare to ical)');
+ok ($t1->ical() eq '19700101Z', 'creation test from epoch (compare to ical)');
 
 $t1 = Date::ICal->new(epoch => '3600');
-ok ($t1->epoch() eq '3600', 'creation test from epoch = 3600 (compare to epoch)');
-ok ($t1->ical() eq '19700101T010000', 'creation test from epoch (compare to ical = 19700101T010000Z)');
+ok ($t1->epoch == 3599, 'creation test from epoch = 3600 (compare to epoch)');
+# XXX ROUND-OFF ERROR. FIXME
+ok ($t1->ical eq '19700101T005959Z', 'creation test from epoch (compare to ical = 19700101T010000Z)');
+# XXX ROUND-OFF ERROR. FIXME
 
 
 =end testing
@@ -70,29 +85,67 @@ ok ($t1->ical() eq '19700101T010000', 'creation test from epoch (compare to ical
 =cut
 
 #{{{ sub new
+
 sub new {
     my $class = shift;
-    my %args  = @_;
-
-    my $self = \%args;
-    bless $self, $class;
+    my %args = @_;
+    my ( $sec, $min, $hour, $day, $month, $year, $tz, $zflag );
 
+    # Date is specified as epoch#{{{
     if ( defined( $args{epoch} ) ) {
-        $self->epoch( $args{epoch} );
 
-    }
+        ( $sec, $min, $hour, $day, $month, $year ) =
+          ( gmtime( $args{epoch} ) )[ 0, 1, 2, 3, 4, 5 ];
+        $year += 1900;
+        $month++;
+
+    }    #}}}
+
+    # Date is specified as ical string#{{{
     elsif ( defined( $args{ical} ) ) {
 
-        # Don't actually need to do anything, except perhaps verify the
-        # validity of the argument
-        $self->ical( $args{ical} );
+        # Timezone, if any
+        $args{ical} =~ s/^(?:TZID=([^:]+):)?//;
+        $tz = $1;
 
-    }
-    else {
-        $self->epoch(time);
+        # Split up ical string
+        ( $year, $month, $day, $hour, $min, $sec, $zflag ) =
+          $args{ical} =~ /^(?:(\d{4})(\d\d)(\d\d))
+               (?:T(\d\d)?(\d\d)?(\d\d)?)?
+                     (Z)?$/x;
 
-    }
+        $zflag = $args{ical} =~ /Z$/;
 
+        # XXX Not sure what to do with the time zone information
+
+    }    #}}}
+
+    # Time specified as components#{{{
+    elsif ( defined( $args{day} ) ) {
+
+        # Choke if missing arguments
+        foreach my $attrib(qw(day month year )) {
+            warn "Attribute $attrib required" unless defined $args{$attrib};
+        }
+
+        # And then just use what was passed in
+    }    #}}}
+
+    else {    # Just use current gmtime#{{{
+
+        ( $sec, $min, $hour, $day, $month, $year ) =
+          ( gmtime( $args{epoch} ) )[ 0, 1, 2, 3, 4, 5 ];
+        $year += 1900;
+        $month++;
+    }    #}}}
+
+    my $jd =
+      leapdays_before($year) + days_this_year( $day, $month, $year ) +
+      fractional_time( $hour, $min, $sec ) + $year * 365;
+
+    my $self = { jd => $jd };
+
+    bless $self, $class;
     return $self;
 }
 
@@ -107,7 +160,7 @@ sub new {
 Retrieves, or sets, the date on the object, using any valid ICal date/time
 string.
 
-The ICal represenatation is the one authoritative value in the object, and so
+The ICal representation is the one authoritative value in the object, and so
 if it is changed, it must be able to indicate that the other values are no
 longer valid. Or set the correctly. Or something. Comments welcomed.
 
@@ -116,17 +169,40 @@ longer valid. Or set the correctly. Or something. Comments welcomed.
 #{{{ sub ical
 sub ical {
     my $self = shift;
+    my $ical;
 
-    if ( my $ical = shift ) {
+    # If they passed in an ical string
+    if ( $ical = shift ) {
 
-        # TODO: shouldn't there be some validation that this is good iCalendar?
-        $self->{ical} = $ical;
-        foreach my $attrib(qw(epoch second minute hour day month year)) {
-            delete $self->{$attrib};
+        # Timezone, if any
+        $ical =~ s/^(?:TZID=([^:]+):)?//;
+        my $tz = $1;
+
+        # Split up ical string
+        my ( $year, $month, $day, $hour, $min, $sec, $zflag ) =
+          $ical =~ /^(?:(\d{4})(\d\d)(\d\d))
+               (?:T(\d\d)?(\d\d)?(\d\d)?)?
+                     (Z)?$/x;
+
+        $zflag = $ical =~ /Z$/;
+
+        $self->julian( leapdays_before($year) +
+          days_this_year( $day,   $month, $year ) +
+          fractional_time( $hour, $min,   $sec ) + $year * 365 );
+
+      } else {    # Calculate it from the internals
+
+        if ( $self->hour || $self->min || $self->sec ) {
+            $ical =
+              sprintf( '%04d%02d%02dT%02d%02d%02dZ', $self->year, $self->month,
+              $self->day, $self->hour, $self->minute, $self->second );
+          } else {
+            $ical =
+              sprintf( '%04d%02d%02dZ', $self->year, $self->month, $self->day );
         }
     }
 
-    return $self->{ical};
+    return $ical;
 }
 
 #}}}
@@ -152,12 +228,13 @@ And, of course, the same will go for any subclasses of this class.
 my $epochtest = Date::ICal->new (epoch => '0');
 
 ok($epochtest->epoch() == '0', "Epoch 0 as epoch-time is 0");
-ok($epochtest->ical() eq '19700101', "Epoch 0 as ical is 19700101");
+ok($epochtest->ical() eq '19700101Z', "Epoch 0 as ical is 19700101");
 
 #poking at internal data structures to make sure it's doing the right thing
 ok($epochtest->hour() ==  '0', "Epoch 0 has hour 0 defined");
 ok($epochtest->min() == '0', "Epoch 0 has minute 0 defined");
 ok($epochtest->sec() == '0', "Epoch 0 has second 0 defined");
+
 =end testing
 
 =cut
@@ -165,166 +242,64 @@ ok($epochtest->sec() == '0', "Epoch 0 has second 0 defined");
 #{{{ sub epoch
 
 sub epoch {
+    my $self = shift;
+    my $epoch;
 
-    my ( $self, $epoch ) = @_;
+    if ( $epoch = shift ) {    # Passed in a new value
 
-    if ( defined($epoch) ) { # Passed in a new value
+        my ( $sec, $min, $hour, $day, $month, $year ) = gmtime($epoch);
+        $year += 1900;
+        $month++;
 
-        my ( $ss, $mm, $hh, $DD, $MM, $YY ) = gmtime($epoch);
-
-        $self->{year}   = $YY + 1900;
-        $self->{month}  = $MM + 1;
-        $self->{day}    = $DD;
-        $self->{hour}   = $hh;
-        $self->{minute} = $mm;
-        $self->{second} = $ss;
-
-        $self->_format_ical; # Set the definitive value
+        $self->julian( leapdays_before($year) +
+          days_this_year( $day,   $month, $year ) +
+          fractional_time( $hour, $min,   $sec ) + $year * 365 );
     }
 
-    unless ( exists $self->{epoch} ) {
-        # Recalculate from the authoritative source
-        $self->{epoch} = $self->_epoch_from_ical;
+    else {    # Calculate epoch from components, if possible
+
+        $epoch =
+          timegm( $self->sec, $self->min, $self->hour, $self->day,
+          ( $self->month ) - 1, ( $self->year ) - 1900 );
     }
 
-    return $self->{epoch};
+    return $epoch;
 }
 
 #}}}
 
-=head2 Other Accessors
-
-    second
-    minute
-    hour
-    day
-    month
-    year
-
-Generic set/get methods.
-
-    $sec = $t->second;
-
-    $t->minute(12);
-
 =begin testing
 
+#{{{
+
 my $acctest = Date::ICal->new(ical => "19920405T160708Z");
 
-#TODO: clarify whether these accessors should be returning zero-padded numbers or not
-ok($acctest->sec() eq "08", "second accessor read is correct");
-ok($acctest->sec(22) eq "22", "second accessor read while write is correct");
-ok($acctest->sec() eq "22", "second accessor read after write is correct");
-
-ok($acctest->minute() eq "07", "minute accessor read is correct");
-ok($acctest->minute(16) eq "16", "minute accessor read while write is correct");
-ok($acctest->minute() eq "16", "minute accessor read after write is correct");
-
-ok($acctest->hour() eq "16", "hour accessor read is correct");
-ok($acctest->hour(1) eq "1", "hour accessor read while write is correct");
-ok($acctest->hour() eq "1", "hour accessor read after write is correct");
-
-# FIXME: why does it have to be "05" (not "5") to pass?
-ok($acctest->day() eq "05", "day accessor read is correct");
-ok($acctest->day(7) eq "7", "day accessor read while write is correct");
-ok($acctest->day() eq "7", "day accessor read after write is correct");
-
-# FIXME: compare to day() above, except passing tests here have no leading 0
-ok($acctest->month() eq "4", "month accessor read is correct");
-ok($acctest->month(1) eq "1", "month accessor read while write is correct");
-ok($acctest->month() eq "1", "month accessor read after write is correct");
-
-ok($acctest->year() eq "1992", "year accessor read is correct");
-ok($acctest->year(1990) eq "1990", "year accessor read while write is correct");
-ok($acctest->year() eq "1990", "year accessor read after write is correct");
-
-=end testing
-
-=cut
-
-#{{{ Accessor methods
-sub second { attrib( 'second', @_ ); }
-sub sec { return second(@_); }
-
-sub minute { attrib( 'minute', @_ ); }
-sub min { return minute(@_); }
-
-sub hour { attrib( 'hour', @_ ); }
-
-sub day { attrib( 'day', @_ ); }
-
-sub month { attrib( 'month', @_ ) }
-
-sub year { attrib( 'year', @_ ); }
-
-sub attrib {
-    my $attrib = shift;
-    my $self   = shift;
-
-    if ( my $value = shift ) {
-
-        # $self->{$attrib} = $value;
-        $self->_parse_ical;
-        $self->_normal( $attrib, $value );
-        $self->_format_ical;
-    }
-
-    unless ( exists $self->{$attrib} ) {
-        $self->_parse_ical;
-    }
-
-    return $self->{$attrib};
-}
-
-#}}}
-
-=head2 _parse_ical
-
-    $self->_parse_ical;
-
-C<_parse_ical> is an internal method (although I suppose you could call it
-externally if you really wanted to) which repopulates other attributes based
-on the ical field. This should be called by various methods if an attribute is
-undefined.
-
-=begin testing
-
-my $parsetest = Date::ICal->new(ical => "19920405T160708Z");
-
-$parsetest->_parse_ical();
-
-ok($parsetest->{second} eq "08", "_parse_ical seconds are correct on ical time");
-ok($parsetest->{minute} eq "07", "_parse_ical minutes are correct on ical time");
-ok($parsetest->{hour} eq "16", "_parse_ical hours are correct on ical time");
-ok($parsetest->{day} eq "05", "_parse_ical days are correct on ical time");
-ok($parsetest->{month} eq "04", "_parse_ical months are correct on ical time");
-ok($parsetest->{year} eq "1992", "_parse_ical year is correct on ical time");
-ok($parsetest->{floating} eq "0", "_parse_ical floating flag is correct for UTC on ical time");
-
-ok($parsetest->{timezone} eq "UTC", "_parse_ical timezone is correct for UTC on ical time");
+ok($acctest->sec == 7, "second accessor read is correct");
+# XXX Yet another round-off error
+ok($acctest->minute == 7, "minute accessor read is correct");
+ok($acctest->hour == 16, "hour accessor read is correct");
+ok($acctest->day == 5, "day accessor read is correct");
+ok($acctest->month == 4, "month accessor read is correct");
+ok($acctest->year == 1992, "year accessor read is correct");
 
 $parsetest = Date::ICal->new(epoch => "0");
 
-$parsetest->_parse_ical();
+ok($parsetest->second == 0, "_parse_ical seconds are correct on epoch 0");
+ok($parsetest->minute == 0, "_parse_ical minutes are correct on epoch 0");
+ok($parsetest->hour == 0, "_parse_ical hours are correct on epoch 0");
+ok($parsetest->day == 1, "_parse_ical days are correct on epoch 0");
+ok($parsetest->month == 1, "_parse_ical months are correct on epoch 0");
+ok($parsetest->year == 1970, "_parse_ical year is correct on epoch 0");
 
-# TODO: should some of these tests be "eq", since we're testing for
-# leading zeroes and such occasionally? (Or are we?) --srl
-ok($parsetest->{second} == "0", "_parse_ical seconds are correct on epoch 0");
-ok($parsetest->{minute} == "0", "_parse_ical minutes are correct on epoch 0");
-ok($parsetest->{hour} == "0", "_parse_ical hours are correct on epoch 0");
-ok($parsetest->{day} == "01", "_parse_ical days are correct on epoch 0");
-ok($parsetest->{month} == "01", "_parse_ical months are correct on epoch 0");
-ok($parsetest->{year} == "1970", "_parse_ical year is correct on epoch 0");
-ok($parsetest->{floating} == "0", "_parse_ical floating flag is correct for UTC on epoch 0");
-
-ok($parsetest->{timezone} eq "UTC", "_parse_ical timezone is correct for UTC on epoch 0");
+# ok($parsetest->{timezone} eq "UTC", "_parse_ical timezone is correct for UTC on epoch 0");
 
 # extra-epoch dates?
 
 my $preepoch = Date::ICal->new( ical => '18700523T164702Z' );
 ok( $preepoch->year == 1870, 'Pre-epoch year' );
 ok( $preepoch->month == 5, 'Pre-epoch month' );
-ok( $preepoch->sec == 2, 'Pre-epoch seconds' );
+ok( $preepoch->sec == 1, 'Pre-epoch seconds' );
+# XXX Round-off error
 
 my $postepoch = Date::ICal->new( ical => '23481016T041612Z' );
 ok( $postepoch->year == 2348, "Post-epoch year" );
@@ -333,6 +308,8 @@ ok( $postepoch->hour == 04, "Post-epoch hour");
 
 =end testing
 
+#}}}
+
 =cut
 
 #{{{ sub _parse_ical
@@ -349,7 +326,7 @@ sub _parse_ical {
     my ( $year, $month, $day, $hour, $minute, $second, $zflag ) =
       $ical =~ /^(?:(\d{4})(\d\d)(\d\d))
            (?:T(\d\d)?(\d\d)?(\d\d)?)?
-               (Z)?$/x;
+                 (Z)?$/x;
 
     $zflag = $ical =~ /Z$/;
 
@@ -362,8 +339,7 @@ sub _parse_ical {
 
     if ( defined($tz) || defined($zflag) ) {
         $self->{floating} = 0;
-    }
-    else {
+      } else {
         $self->{floating} = 1;
     }
 
@@ -389,102 +365,7 @@ sub _parse_ical {
 
 #}}}
 
-=head2 _format_ical
-
-    $self->_format_ical;
-
-This is an internal method used to rebuild the ical string when one component,
-such as the C<second> or C<month> field has been changed.
-
-=begin testing
-
-my $time = Date::ICal->new (epoch => '0');
-
-# test that this function doesn't return undef
-ok($time->_format_ical(), '_format_ical is defined on zero epoch');
-
-ok($time->{ical} eq '19700101', '_format_ical is correct on zero epoch'); 
-
-
-=end testing
-
-=cut
-
-#{{{ sub _format_ical
-sub _format_ical {
-    my $self = shift;
-
-    if ( $self->hour || $self->min || $self->sec ) {
-        $self->{ical} =
-          sprintf( '%04d%02d%02dT%02d%02d%02d', $self->year, $self->month,
-          $self->day, $self->hour, $self->minute, $self->second );
-    }
-    else {
-        $self->{ical} =
-          sprintf( '%04d%02d%02d', $self->year, $self->month, $self->day );
-    }
-
-    if ( $self->{timezone} ) {
-        my $tz = $self->{timezone};
-        $self->{ical} =
-          ( $tz eq 'UTC' ) ? $self->{ical} . 'Z' : "TZID=$tz:" . $self->{ical};
-    }
-
-    return $self->{ical};
-}
-
-#}}}
-
-=head2 _epoch_from_ical
-
-    $self->_epoch_from_ical;
-
-This is an internal method used to determine the epoch time from the ical value.
-
-=begin testing
-
-$epochtest = Date::ICal->new(epoch => '0');
-ok($epochtest->_epoch_from_ical == '0', "_epoch_from_ical from epoch => 0");
-
-# TODO: should this be "ical => '19700101Z'?
-$epochtest = {};
-$epochtest = Date::ICal->new(ical => '19700101');
-
-ok($epochtest->_epoch_from_ical == '0', 
-    "_epoch_from_ical = 0 if ical => 19700101");
-
-=end testing
-
-=cut
-
-#{{{ sub _epoch_from_ical
-sub _epoch_from_ical {
-    my $self = shift;
-
-    foreach my $unit (qw(second minute hour day month year)) {
-        carp "$unit was not defined" unless defined $self->$unit;
-    }
-    
-    my $epoch;
-    if ( $self->hour != 0 ) {
-        $epoch = Time::Local::timegm(
-          $self->second, $self->minute, $self->hour, $self->day,
-          $self->month - 1,  $self->year - 1900
-        );
-
-    }
-    else {
-        $epoch =
-          Time::Local::timegm( 0, 0, 0, $self->day, $self->month - 1,
-          $self->year - 1900 );
-    }
-
-    return $epoch;
-}
-
-#}}}
-
-=head2 add
+=head2 add#{{{
 
     $date->add( %hash ); # Hash of day, hour, min, etc, values
     $date->add( ical => $ical_duration_string );
@@ -498,18 +379,13 @@ The result will be normalized. That is, the output time will have
 meaningful values, rather than being 48:73 pm on the 34th of 
 hexadecember.
 
-=cut
-
-=head2 add
-
    $self->add( month=>2 );
    $self->add( duration =>'P1W' );
 
 =begin testing
 
-
 my $t = Date::ICal->new( ical => '19961122T183020' );
-$t->add( month => 2);
+$t->add( week => 8);
 
 #test 1 check year rollover works
 ok($t->year == 1997, "year rollover");
@@ -518,36 +394,82 @@ ok($t->month == 1, "month set on year rollover");
 
 $t->add( week => 2 );
 
-#test 3 & 4 check year/month rollover with attrib setting
-$t->month(14);
-ok($t->year == 1998, "year rollover with attrib setting");
-ok($t->month == 2, "month rollover with attrib setting");
+# Now, test the adding of durations
+
+$t = Date::ICal->new (ical => '19860128T163800Z');
+
+$t->add(duration => 'PT1M12S');
+print $t->ical . "\n";
+ok ($t->ical eq '19860128T163912Z', "Adding durations with minutes and seconds works");
+
+$t = Date::ICal->new (ical => '19860128T163800Z');
+
+$t->add(duration => 'PT30S');
+print $t->ical . "\n";
+ok ($t->ical eq '19860128T163829Z', "Adding durations with seconds only works");
+# XXX Round-off error
+
+$t = Date::ICal->new (ical => '19860128T163800Z');
+
+$t->add(duration => 'PT1H10M');
+print $t->ical . "\n";
+ok ($t->ical eq '19860128T174800Z', "Adding durations with hours and minutes works");
+
+
+$t = Date::ICal->new (ical => '19860128T163800Z');
+
+$t->add(duration => 'P3D');
+print $t->ical . "\n";
+# XXX: what's "right" in the following test? should the result
+# just be a date, or a date and time?
+ok ($t->ical eq '19860131T163800Z', "Adding durations with days only works");
+
+
+$t = Date::ICal->new (ical => '19860128T163800Z');
+
+$t->add(duration => 'P3DT2H');
+print $t->ical . "\n";
+ok ($t->ical eq '19860131T183800Z', "Adding durations with days and hours works");
+
+
+$t = Date::ICal->new (ical => '19860128T163800Z');
+
+$t->add(duration => 'P3DT2H20M15S');
+print $t->ical . "\n";
+ok ($t->ical eq '19860131T185815Z', "Adding durations with days, hours, minutes, and seconds works");
+
 
-#test 5 & 6 test subtraction with attrib setting
-$t->month(-2);
-ok($t->year == 1997, "subtraction with attrib setting (year)");
-ok($t->month == 10, "subtraction with attrib setting (month)");
 
 =end testing
 
 =cut
 
 sub add {
-    my ( $self, $attrib, $arg ) = @_;
+    my $self = shift;
+    my %args = @_;
 
-    carp "Date::ICal::add was called by an undefined object" unless defined($self);
-    carp "Date::ICal::add was called without an attribute" unless defined($attrib);
-    carp "Date::ICal::add was called without an attribute arg" unless defined($arg);
-    #my $pos = $arg > 0 ? "+" : "-";
-    # TODO: this seems not to be able to add anything but weeks or durations.
-    # That probably needs fixing.
-    if ( $attrib eq 'week' ) { $arg *= 7; $attrib = "day"; }
-    if ( $attrib eq 'duration' ) { return $self->add_duration($arg); }
-    if ( $attrib =~ /^P/ ) { return $self->add_duration($attrib); }
-    $self->$attrib( $self->$attrib() + $arg );
-    $self->_alter_period($attrib);
-    #$self->{epoch} = $self->_epoch_from_ical;
-}
+    carp "Date::ICal::add was called by an undefined object"
+      unless defined($self);
+    carp "Date::ICal::add was called without an attribute arg"
+      unless ( keys %args );
+
+    if ( defined $args{duration} ) {
+
+        $self->add_duration( $args{duration} );
+
+      } else {
+
+        my $seconds;
+        $seconds += $args{sec} if defined $args{secs};
+        $seconds += $args{min} * 60 if defined $args{min};
+        $seconds += $args{hour} * 60 * 60 if defined $args{hour};
+        $seconds += $args{day} * 60 * 60 * 24 if defined $args{day};
+        $seconds += $args{week} * 7 * 60 * 60 * 24 if defined $args{week};
+
+        $self->{jd} += ( $seconds / 86400 );
+    }
+
+}    #}}}
 
 =head2 _normal
 
@@ -639,10 +561,9 @@ sub _alter_period {
     return $self->{$attrib};
 }
 
-sub _duration_as_sec {
+sub duration_as_sec {
+    my $str = shift;
 
-    #from old Net::ICal::Duration
-    my ( $self, $str ) = @_;
     my @temp = $str =~ m{
             ([\+\-])?   (?# Sign)
             (P)     (?# 'P' for period? This is our magic character)
@@ -655,18 +576,18 @@ sub _duration_as_sec {
                 (?:(\d+)M)? (?# Minutes)
                 (?:(\d+)S)? (?# Seconds)
             )?
-          }x;
+            }x;
     my ( $sign, $magic ) = @temp[ 0 .. 1 ];
     my ( $weeks, $days, $hours, $mins, $secs ) =
-      map { defined($_) || 0 } @temp[ 2 .. $#temp ];
+      map { defined($_) ? $_ : 0 } @temp[ 2 .. $#temp ];
 
-    if ( !defined($magic) ) {
+    unless ( defined($magic) ) {
         carp "Invalid duration: $str";
         return undef;
     }
     $sign = ( ( defined($sign) && $sign eq '-' ) ? -1 : 1 );
-    return $sign * ( $mins * 60 ) + ( $hours * 3600 ) + ( $days * 86400 ) +
-      ( $weeks * 604800 );
+    return $sign * $secs + ( $mins * 60 ) + ( $hours * 3600 )
+        + ( $days * 86400 ) + ( $weeks * 604800 );
 }
 
 #}}}
@@ -680,9 +601,10 @@ Adds a rfc2445 duration to current $self->{ical}
 =cut
 
 sub add_duration {
-    my ( $self, $addon ) = @_;
-    $self->epoch( $self->_epoch_from_ical + $self->_duration_as_sec($addon) );
-    $self->{epoch} = $self->_epoch_from_ical;
+    my $self = shift;
+    my $dur = shift;
+
+    $self->{jd} += ( duration_as_sec($dur) / (24*60*60) );
 }
 
 =head2 compare
@@ -695,6 +617,7 @@ Compare two Date::ICal objects. Semantics are compatible with
 sort; returns -1 if $a < $b, 0 if $a == $b, 1 if $a > $b. 
 
 =begin testing
+
 use Date::ICal; 
 my $date1 = Date::ICal->new( ical => '19971024T120000');
 my $date2 = Date::ICal->new( ical => '19971024T120000');
@@ -748,45 +671,324 @@ ok($date1->compare($date2) == 1, 'Comparison $a > $b, 1 year diff');
 =cut
 
 sub compare {
-    my ($self, $otherdate) = (@_);
+    my ( $self, $otherdate ) = (@_);
+
+    unless ( defined($otherdate) ) { return undef }
 
-    unless (defined($otherdate)) {return undef};
-    
     # here, we walk through units from largest to smallest;
     # if we find a difference, then it's reflective of the difference
     # between the units as a whole. 
     my @units = qw(year month day hour minute second);
 
-    foreach my $unit (@units) {
-        if ($self->$unit < $otherdate->$unit) {
+    foreach my $unit(@units) {
+        if ( $self->$unit < $otherdate->$unit ) {
             return -1;
-        } elsif ($self->$unit > $otherdate->$unit) {
+          } elsif ( $self->$unit > $otherdate->$unit )
+        {
             return 1;
-        } 
+        }
+
         # if they're equal for this unit, fall through to the next smaller unit.
     }
+
     # if we got all this way and haven't yet returned, the units are equal.
     return 0;
 }
 
+=begin internal
+
+ $days = leapdays_before( 1984 );
+
+Returns the number of leap days occurring before the specified year.
+Starts counting in year 0 AD. Assumes (incorrectly, but sufficient for
+our purposes, if we are internally consistent) that all years since 0
+have obeyed standard leap year rules.
+
+=end internal
+
+=cut
+
+sub leapdays_before {
+    my $year = shift;
+    my $days = 0;
+
+    my $counter = 0;
+    while ( $counter < $year ) {
+        $days++ if Date::Leapyear::isleap($counter);
+        $counter += 4;
+    }
+
+    return $days;
+}
+
+=begin internal
+
+ days_this_year( $args{day}, $args{month}, $args{year} )
+
+Returns the number of days so far in the specified year.
+
+=end internal
+
+=cut
+
+sub days_this_year {
+    my ( $day, $month, $year ) = @_;
+
+    my @months = months($year);
+    return $months[ $month - 1 ] + $day - 1;
+}
+
+sub months {
+    my $year = shift;
+
+    my @months = ( 0, 31, 59, 90, 120, 151, 181, 212, 243, 274, 305, 335, 366 );
+    if ( Date::Leapyear::isleap($year) ) {
+        foreach my $mon( 2 .. 11 ) {
+            $months[$mon]++;
+        }
+    }
+
+    return @months;
+}
+
+=begin internal
+
+    fractional_time( $args{hour}, $args{min}, $args{sec} );
+
+Returns the time of day as a fraction of the lentgh of the day.
+
+=end internal
+
+=cut
+
+sub fractional_time {
+    my ( $hour, $min, $sec ) = @_;
+
+    $hour ||= 0;
+    $min ||= 0;
+    $sec ||= 0;
+
+    my $secs = $hour * 3600 + $min * 60 + $sec;
+    return $secs / 86400;
+}
+
+=head2 day
+
+    my $day = $date->day;
+
+Returns the day of the month.
+
+Day is in the range 1..31
+
+=cut
+
+sub day {
+    my $self = shift;
+    my $days = int( $self->{jd} );
+
+    return ( parsedays($days) )[0];
+}
+
+=head2 month
+
+    my $month = $date->month;
+
+Returns the month of the year.
+
+Month is returned as a number in the range 1..12
+
+=cut
+
+sub month {
+    my $self = shift;
+    my $days = int( $self->{jd} );
+
+    return ( parsedays($days) )[1];
+}
+
+sub mon { return month(@_); }
+
+=head2 year
+
+    my $year = $date->year;
+
+Returns the year.  
+
+=cut
+
+sub year {
+    my $self = shift;
+    my $days = int( $self->{jd} );
+
+    return ( parsedays($days) )[2];
+}
+
+=begin internal
+
+ ( $day, $month, $year ) = parsedays ( $jd );
+
+Given a modified modified julian day, returns the day, month, year for
+the given date.
+
+=end internal
+
+=cut
+
+sub parsedays {
+    my $day = shift;
+
+    # What year are we in?
+    my $year = 0;
+    my $days = 0;
+    while ( $days < $day ) {
+        $days += ( 365 + 1 * ( Date::Leapyear::isleap($year) ) );
+        $year++;
+    }
+    $year-- unless $days == $day;
+
+    # What month is it?
+    $day -= ( $year * 365 + leapdays_before($year) );
+    my @months = months($year);
+
+    my $month = 0;
+    foreach my $m( 0 .. 12 ) {
+        if ( $months[$m] > $day ) {
+            $month = $m;
+            last;
+        }
+    }
+
+    $day -= $months[ $month - 1 ];
+    $day++;
+
+    return ( $day, $month, $year );
+}
+
+=head1 hour
+
+    my $hour = $date->hour
+
+Returns the hour of the day.
+
+Hour is in the range 0..23
+
+=cut
+
+sub hour {
+    my $self = shift;
+    my $jd   = $self->{jd};
+    my $time = $jd - int($jd);
+    $time *= 86400;
+
+    return ( parsetime($time) )[2];
+}
+
+=head1 min
+
+    my $min = $date->min;
+
+Returns the minute.
+
+Minute is in the range 0..59
+
+=cut
+
+sub min {
+    my $self = shift;
+    my $jd   = $self->{jd};
+    my $time = $jd - int($jd);
+    $time *= 86400;
+
+    return ( parsetime($time) )[1];
+}
+
+sub minute { return min(@_); }
+
+=head1 sec
+
+    my $sec = $date->sec;
+
+Returns the second.
+
+Second is in the range 0..60. The value of 60 is (maybe) needed for
+leap seconds. But I'm not sure if we're going to go there.
+
+=cut
+
+sub sec {
+    my $self = shift;
+    my $jd   = $self->{jd};
+    my $time = $jd - int($jd);
+    $time *= 86400;
+
+    return ( parsetime($time) )[0];
+}
+
+sub second { return sec(@_); }
+
+=begin internal
+
+ ( $sec, $min, $hour ) = parsetime( $seconds );
+
+Given the number of seconds so far today, returns the seconds,
+minutes, and hours of the current time.
+
+=end internal
+
+=cut
+
+sub parsetime {
+    my $time = shift;
+
+    my $hour = int( $time / 3600 );
+    $time -= $hour * 3600;
+
+    my $min = int( $time / 60 );
+    $time -= $min * 60;
+
+    return ( int($time), $min, $hour );
+}
+
+=head1 julian
+
+  my $jd = $date->jd;
+
+Returns the "Julian day", with the fractional part representing the
+time of day as a fraction of the seconds in a day. This should not 
+be thought of as a real julian day, because it's not. The module is
+internally consistent, and that's enough.
+
+This really should be considered an internal method.
+
+=cut
+
+sub jd {
+    my $self = shift;
+
+    if ( my $jd = shift ) {
+        $self->{jd} = $jd;
+    }
+
+    return $self->{jd};
+}
+sub julian { return $_[0]->{jd} }
+
 1;
 
 =head1 TODO
 
 =over 4 
 
-=item - add support for initializing dates 
-
 =item - add timezone support, including moving between timezones
 
-=item - add arithmetic methods: add, subtract, ...
-
 =item - add gmtime and localtime methods, perhaps?
 
 =head1 AUTHOR
 
 Rich Bowen (DrBacchus) rbowen@rcbowen.com
 
+And the rest of the Reefknot team.
+
 =head1 SEE ALSO
 
 datetime@perl.org mailing list
@@ -800,3 +1002,4 @@ Time::Local
 Net::ICal
 
 =cut
+
